<?php

namespace Drupal\custom_elements_ui\Form;

use Drupal\Component\Plugin\Factory\DefaultFactory;
use Drupal\Component\Plugin\PluginManagerBase;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityWithPluginCollectionInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Field\PluginSettingsInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Url;
use Drupal\custom_elements\CustomElementsFieldFormatterInterface;
use Drupal\custom_elements\CustomElementsFieldFormatterPluginManager;
use Drupal\custom_elements\Entity\EntityCeDisplayInterface;
use Drupal\field_ui\FieldUI;
use Drupal\field_ui\Form\EntityDisplayFormBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Forked version of EntityDisplayFormBase class of core.
 *
 * The EntityCeDisplay object has makes use of CeFormatters instead of core
 * field formatters for rendering display components - as returned by
 * EntityCeDisplay::getRenderer(). Our CE-field-formatters are properly
 * configurable plugins. So instead of having plugin-settings forms like
 * core-field formatters, we have plugin configuration forms. The existing
 * EntityDisplay logic has been adapted to the new API, some keys, variable
 * names and comments still refer to "plugin settings" instead of "plugin
 * configuration" though.
 * This is good to keep, because we are also supporting the configuration of
 * core field formatters as part of our CoreFieldCeFieldFormatter plugins.
 *
 * However, the core field formatter settings form API is simple and pre-dates
 * the core SubformState, so it needs to have the same global location in the
 * form array as core does it. Long story short, we need to keep the
 * 'settings_edit_form' key there. For example,
 * empty_fields_field_formatter_third_party_settings_form() relies on it.
 *
 * @internal
 *   Form classes are internal.
 */
class EntityCustomElementsDisplayEditForm extends EntityDisplayFormBase {

  /**
   * {@inheritdoc}
   */
  protected $displayContext = 'view';

  /**
   * The custom element formatter plugin manager.
   *
   * @var \Drupal\custom_elements\CustomElementsFieldFormatterPluginManager
   */
  protected $customElementFormatterPluginManager;

  /**
   * Constructs a new EntityCustomElementsDisplayEditForm.
   *
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $field_type_manager
   *   The field type manager.
   * @param \Drupal\Component\Plugin\PluginManagerBase $plugin_manager
   *   The widget or formatter plugin manager.
   * @param \Drupal\Core\Entity\EntityDisplayRepositoryInterface $entity_display_repository
   *   (optional) The entity display_repository.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   (optional) The entity field manager.
   * @param \Drupal\custom_elements\CustomElementsFieldFormatterPluginManager $ce_formatter_plugin_manager
   *   The custom element formatter plugin manager.
   */
  public function __construct(
    FieldTypePluginManagerInterface $field_type_manager,
    PluginManagerBase $plugin_manager,
    EntityDisplayRepositoryInterface $entity_display_repository,
    EntityFieldManagerInterface $entity_field_manager,
    CustomElementsFieldFormatterPluginManager $ce_formatter_plugin_manager,
  ) {
    parent::__construct($field_type_manager, $plugin_manager, $entity_display_repository, $entity_field_manager);
    $this->customElementFormatterPluginManager = $ce_formatter_plugin_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('plugin.manager.field.field_type'),
      $container->get('plugin.manager.field.formatter'),
      $container->get('entity_display.repository'),
      $container->get('entity_field.manager'),
      $container->get('custom_elements.plugin.manager.field.custom_element_formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\custom_elements\Entity\EntityCeDisplayInterface $entity */
    $entity = $this->getEntity();

    $form['force_auto'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Automatic processing'),
      '#description' => $this->t('Skip this configuration screen; let Drupal figure out how to build the custom element.'),
      '#default_value' => $entity->getForceAutoProcessing(),
    ];

    $form['custom_element_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Custom element name'),
      '#description' => $this->t('Defines the custom element which is going to be used for the display. There must be a matching implementation in the frontend to render the display.'),
      '#default_value' => $entity->getCustomElementName(),
      '#size' => 20,
      '#required' => TRUE,
    ];
    $layout_builder_enabled = $this->isLayoutBuilderEnabled();
    if ($layout_builder_enabled) {
      // Name is not used, only if Layout Builder is enabled and forceAuto = on.
      // @todo Make the name not required when layout-builder=on and auto=on?
      $form['custom_element_name']['#states'] = [
        'visible' => [
          ':input[name="force_auto"]' => ['checked' => FALSE],
        ],
      ];

      // @todo We also need to make this invisible (and the field rows visible)
      //   if auto=off. But first, do #3446485.
      $display = $this->entityDisplayRepository->getViewDisplay($entity->getTargetEntityTypeId(), $entity->getTargetBundle(), $entity->getMode());
      $entity_type = $this->entityTypeManager->getDefinition($display->getTargetEntityTypeId());
      $bundle_parameter_key = $entity_type->getBundleEntityType() ?: 'bundle';
      $route_parameters = [
        $bundle_parameter_key => $display->getTargetBundle(),
        'view_mode_name' => $display->getMode(),
      ];

      $form['help'] = [
        '#type' => 'markup',
        '#markup' => $this->t('<p><em>Layout builder</em> is enabled for this view mode. Thus, for this view mode the configured layout sections and the contained blocks are getting rendered as custom element. Configure the layout to customize how the element should be built.</p>'),
      ];

      $form['manage_layout'] = [
        '#type' => 'link',
        '#title' => $this->t('Manage layout'),
        '#attributes' => ['class' => ['button']],
        '#url' => Url::fromRoute("layout_builder.defaults." . $entity->getTargetEntityTypeId() . '.view', $route_parameters),
      ];
    }

    // When displaying the 'default' view mode and layout builder is enabled,
    // the form must still display/save the checkboxes for enabling view modes.
    // @todo test: In this case, do not save the CE entity at all?
    if (!$layout_builder_enabled || $entity->getMode() === 'default') {
      if (!$layout_builder_enabled) {
        $form['help'] = [
          '#type' => 'markup',
          '#markup' => $this->t("<p><strong>Warning: this user interface is subject to change and contains some bugs. As a result:</strong><ul><li><strong>If you want to rename a 'Key': immediately afterwards, press TAB and then press Save for each individual renamed key.</strong></li><li><strong>When editing anything else, do not rename keys at the same time.</strong></li></ul></p>"),
        ];
      }
      $form = parent::form($form, $form_state);

      if ($layout_builder_enabled) {
        unset($form['fields'], $form['refresh']);
      }
      else {
        // @todo hide $form['fields'] or equivalent if autoprocessing is on. (It
        //   does not listen to #states at the moment.) Not likely to be done
        //   before #3446485 == reorganizing the whole section.
        // Remove "extra fields".
        foreach (array_keys($this->getExtraFields()) as $field_id) {
          unset($form['fields'][$field_id]);
        }
      }
    }

    // Remove EntityForm::afterBuild(), which calls copyFormValuesToEntity();
    // we cannot update $this->entity before form values are properly validated.
    // Any other validation functions must take this into account.
    if (!empty($form['#after_build'])) {
      $form['#after_build'] = array_diff($form['#after_build'], ['::afterBuild']);
    }

    return $form;
  }

  /**
   * Checks layout_builder is enabled for the current view display.
   *
   * @return bool
   *   TRUE if it's enabled FALSE otherwise.
   */
  protected function isLayoutBuilderEnabled() {
    /** @var \Drupal\custom_elements\Entity\EntityCeDisplayInterface $entity */
    $entity = $this->getEntity();
    $display = $this->entityDisplayRepository->getViewDisplay($entity->getTargetEntityTypeId(), $entity->getTargetBundle(), $entity->getMode());
    return $display->getThirdPartySetting('layout_builder', 'enabled');
  }

  /**
   * {@inheritdoc}
   */
  protected function getEntityDisplay($entity_type_id, $bundle, $mode) {
    $entity_ce_display_storage = $this->entityTypeManager->getStorage('entity_ce_display');
    $entity_ce_display_id = $entity_type_id . '.' . $bundle . '.' . $mode;
    $entity_ce_display = $entity_ce_display_storage->load($entity_ce_display_id);
    // Create new display when opening the form the first time.
    if (empty($entity_ce_display)) {
      /** @var \Drupal\custom_elements\Entity\EntityCeDisplay $entity_ce_display */
      $entity_ce_display = $entity_ce_display_storage->create([
        'targetEntityType' => $entity_type_id,
        'bundle' => $bundle,
        'mode' => $mode,
      ]);
    }
    return $entity_ce_display;
  }

  /**
   * {@inheritdoc}
   */
  protected function getDefaultPlugin($field_type) {
    return 'auto';
  }

  /**
   * Builds the table row structure for a single field.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   A table row array.
   *
   * @todo Redo this with UI revamp in #3455435. parent:form() currently loops
   *   through field definitions, but we likely want to loop through component
   *   names instead, so that multiple components for the same fields can be
   *   set. This means the parameter defenition of buildFieldRow() likey isn't
   *   good to use anymore.
   * @todo At the same time (or later?), decide whether we can support other
   *   'kinds of formatters' (like static values, #3446287) in that same loop.
   *   If not, we'll need to create some separate loop later (just like
   *   parent::form() loops through the extraFields at the moment). Ordering
   *   likely doesn't matter much, since this is going to be reordered by
   *   weight later, anyway.
   * @todo Doublecheck whether the buildExtraFieldRow() calls in parent::form()
   *   do anything, and whether they can be removed.
   */
  protected function buildFieldRow(FieldDefinitionInterface $field_definition, array $form, FormStateInterface $form_state) {
    $display_options = NULL;
    $field_name = $field_definition->getName();
    $label = $field_definition->getLabel();
    $default_name = str_starts_with($field_name, 'field_') ? substr($field_name, strlen('field_')) : $field_name;
    $component_name = $this->getComponentNameFromFieldName($field_name, NULL, FALSE);
    if ($component_name) {
      $display_options = $this->entity->getComponent($component_name);

      // Disable fields without any applicable plugins.
      if (empty($this->getApplicablePluginOptions($field_definition))) {
        $this->entity->removeComponent($component_name);
        $display_options = $this->entity->getComponent($component_name);
      }
    }

    $field_row['human_name'] = [
      '#plain_text' => $label,
    ];
    $field_row['name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Attribute / Slot name'),
      '#title_display' => 'invisible',
      '#default_value' => $component_name ?: $default_name,
      '#size' => 20,
      '#required' => empty($display_options) || $display_options['region'] != 'hidden',
    ];
    $field_row['is_slot'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Is Slot'),
      '#title_display' => 'invisible',
      '#default_value' => $display_options['is_slot'] ?? FALSE,
    ];
    $field_row['formatter'] = [
      '#type' => 'select',
      '#title' => $this->t('Custom element formatter'),
      '#options' => $this->getApplicablePluginOptions($field_definition),
      '#title_display' => 'invisible',
      '#default_value' => $display_options['formatter'] ?? 'auto',
    ];
    $field_row += [
      '#attributes' => ['class' => ['draggable', 'tabledrag-leaf']],
      '#row_type' => 'field',
      '#region_callback' => [$this, 'getRowRegion'],
      '#js_settings' => [
        'rowHandler' => 'field',
        'defaultPlugin' => $this->getDefaultPlugin($field_definition->getType()),
      ],
      'weight' => [
        '#type' => 'textfield',
        '#title' => $this->t('Weight for @title', ['@title' => $label]),
        '#title_display' => 'invisible',
        '#default_value' => $display_options ? $display_options['weight'] : '0',
        '#size' => 3,
        '#attributes' => ['class' => ['field-weight']],
      ],
      'parent_wrapper' => [
        // Somehow FieldUiTable element needs this, so keep it.
        'parent' => [
          '#type' => 'hidden',
          '#empty_value' => '',
          '#attributes' => ['class' => ['js-field-parent', 'field-parent']],
          '#parents' => ['fields', $field_name, 'parent'],
        ],
        'hidden_name' => [
          '#type' => 'hidden',
          '#default_value' => $field_name,
          '#attributes' => ['class' => ['field-name']],
        ],
        'existing_name' => [
          '#type' => 'hidden',
          '#default_value' => $component_name,
        ],
      ],
      'region' => [
        '#type' => 'select',
        '#title' => $this->t('Region for @title', ['@title' => $label]),
        '#title_display' => 'invisible',
        '#options' => $this->getRegionOptions(),
        '#default_value' => $display_options ? $display_options['region'] : 'hidden',
        '#attributes' => ['class' => ['field-region']],
      ],
    ];

    // Get the corresponding plugin object.
    // @todo Safely remove the case for !$component_name in #3446485 / second
    //   call parameter, when all rows represent components.
    $plugin = $component_name ? $this->entity->getRenderer($component_name)
      : $this->entity->getRenderer($field_name, TRUE);

    if ($plugin) {
      $field_row = $this->buildFieldRowPluginForm($field_row, $field_name, $plugin, $form, $form_state);
    }
    return $field_row;
  }

  /**
   * Builds the plugin form into the field row.
   *
   * @param array $field_row
   *   The field form row.
   * @param string $field_name
   *   The namae of the field.
   * @param \Drupal\custom_elements\CustomElementsFieldFormatterInterface $plugin
   *   The formatter plugin.
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return array
   *   The update field form row.
   */
  private function buildFieldRowPluginForm(array $field_row, string $field_name, CustomElementsFieldFormatterInterface $plugin, array $form, FormStateInterface $form_state) {
    // Base button element for the various plugin settings actions.
    $base_button = [
      '#submit' => ['::multistepSubmit'],
      '#ajax' => [
        'callback' => '::multistepAjax',
        'wrapper' => 'field-display-overview-wrapper',
        'effect' => 'fade',
      ],
      '#field_name' => $field_name,
    ];

    $field_row['plugin']['settings_edit_form']['form'] = [];
    $subform_state = SubformState::createForSubform($field_row['plugin']['settings_edit_form']['form'], $form, $form_state);
    $field_row['plugin']['settings_edit_form']['form'] = $plugin
      ->buildConfigurationForm($field_row['plugin']['settings_edit_form']['form'], $subform_state);

    if ($form_state->get('plugin_settings_edit') == $field_name) {
      // We are currently editing this field's plugin settings. Display the
      // settings form and submit buttons.
      if ($field_row['plugin']['settings_edit_form']['form']) {
        $field_row['plugin']['#cell_attributes'] = ['colspan' => 3];
        $field_row['plugin']['settings_edit_form'] = [
          '#type' => 'container',
          '#attributes' => ['class' => ['field-plugin-settings-edit-form']],
          '#parents' => ['fields', $field_name, 'settings_edit_form'],
          'label' => [
            '#markup' => $this->t('Plugin settings'),
          ],
          'form' => $field_row['plugin']['settings_edit_form']['form'],
          'actions' => [
            '#type' => 'actions',
            'save_settings' => $base_button + [
              '#type' => 'submit',
              '#validate' => ['::validatePluginConfigurationForm'],
              '#button_type' => 'primary',
              '#name' => $field_name . '_plugin_settings_update',
              '#value' => $this->t('Update'),
              '#op' => 'update',
            ],
            'cancel_settings' => $base_button + [
              '#type' => 'submit',
              '#name' => $field_name . '_plugin_settings_cancel',
              '#value' => $this->t('Cancel'),
              '#op' => 'cancel',
                // Do not check errors for the 'Cancel' button, but make sure we
                // get the value of the 'plugin type' select.
              '#limit_validation_errors' => [['fields', $field_name, 'type']],
            ],
          ],
        ];
        $field_row['#attributes']['class'][] = 'field-plugin-settings-editing';
      }
    }
    else {
      $field_row['settings_summary'] = [];
      $field_row['settings_edit'] = [];

      // Display a summary of the current plugin settings, and (if the
      // summary is not empty) a button to edit them.
      $summary = $plugin->settingsSummary();

      if (!empty($summary)) {
        $field_row['settings_summary'] = [
          '#type' => 'inline_template',
          '#template' => '<div class="field-plugin-summary">{{ summary|safe_join("<br />") }}</div>',
          '#context' => ['summary' => $summary],
          '#cell_attributes' => ['class' => ['field-plugin-summary-cell']],
        ];
      }

      // When the plugin provides a form, show an edit button.
      // But actually hide the form for now.
      if (!empty($field_row['plugin']['settings_edit_form']['form'])) {
        unset($field_row['plugin']['settings_edit_form']['form']);
        $field_row['settings_edit'] = $base_button + [
          '#type' => 'image_button',
          '#name' => $field_name . '_settings_edit',
          '#src' => 'core/misc/icons/787878/cog.svg',
          '#attributes' => ['class' => ['field-plugin-settings-edit'], 'alt' => $this->t('Edit')],
          '#op' => 'edit',
            // Do not check errors for the 'Edit' button, but make sure we get
            // the value of the 'plugin formatter' select.
          '#limit_validation_errors' => [['fields', $field_name, 'formatter']],
          '#prefix' => '<div class="field-plugin-settings-edit-wrapper">',
          '#suffix' => '</div>',
        ];
      }
    }
    return $field_row;
  }

  /**
   * Gets component name from field name.
   *
   * This is temporary code as long as the UI still has one row per field. It
   * doesn't check whether the CE display has multiple components with the
   * same field name - which isn't supported by the UI yet. There is also no
   * check on whether a CE display has 'non-field' components.
   *
   * WARNING: Re-saving such CE displays through the UI will mess up data!
   *
   * @return string
   *   The component name.
   */
  private function getComponentNameFromFieldName(string $field_name, $entity = NULL, $log = TRUE): string {
    if (!isset($entity)) {
      $entity = $this->entity;
    }
    $component_name = '';
    $components = $entity->getComponents();
    foreach ($components as $name => $component) {
      if (isset($component['field_name']) && $component['field_name'] == $field_name) {
        $component_name = $name;
      }
    }

    if (!$component_name && $log) {
      $this->logger('custom_elements')->warning('No component name found for field @field_name. Alpha UI code is unstable.', ['@field_name' => $field_name]);
    }

    return $component_name;
  }

  /**
   * Form validation callback for the full form.
   *
   * @param array $form
   *   A nested array of form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);
    $form_values = $form_state->getValues();

    // Validate that no duplicate component name is used.
    // @todo there still is a bug with this, as long as the 'name' element
    //   still has AJAX behavior / triggers requests while renaming individual
    //   duplicate elements. Fix (remove AJAX behavior) in #3446485.
    $component_names_used = [];
    foreach ($form['#fields'] as $field_name) {
      $name = $form_values['fields'][$field_name]['name'];
      if (isset($component_names_used[$name])) {
        $form_state->setError($form['fields'][$field_name]['name'], $this->t('Duplicate name %name', ['%name' => $name]));
      }
      $component_names_used[$name] = TRUE;
    }

    // If the form isn't being submitted yet: need to set all components in
    // the form's entity, so all plugin forms are rebuilt correctly.
    if (!$form_state->getErrors() && !empty($form_state->getTriggeringElement()['#ajax'])) {
      $this->doCopyFormValuesToEntity($this->entity, $form, $form_state);
    }
  }

  /**
   * Form validation callback for plugin configuration form.
   *
   * @param array $form
   *   A nested array of form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function validatePluginConfigurationForm(array &$form, FormStateInterface $form_state) {
    // The entity was not populated yet on afterBuild(), because we cannot
    // guarantee correctness of all components when they are being renamed. So,
    // temporarily set current component into cloned entity to make sure we can
    // at least validate its configuration form.
    $validate_entity = clone $this->entity;
    $field_name = $form_state->getTriggeringElement()['#field_name'];
    $component_name = $this->getComponentNameFromFieldName($field_name, $validate_entity);
    $this->doCopyFormValuesToEntity($validate_entity, $form, $form_state, $component_name);

    // Allow the 'triggering' formatter to validate its configuration.
    $plugin = $validate_entity->getRenderer($component_name);
    if ($plugin) {
      $plugin_form =& $form['fields'][$field_name]['plugin']['settings_edit_form']['form'];
      $subform_state = SubformState::createForSubform($plugin_form, $form, $form_state);
      $plugin->validateConfigurationForm($plugin_form, $subform_state);
    }
    else {
      // Should never happen. Not a user-friendly message.
      $form_state->setError($form['fields'][$field_name]['name'], $this->t('Cannot find renderer for component @name', ['@name' => $component_name]));
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    // See form(): this is only called on submit, with form values validated.
    $this->doCopyFormValuesToEntity($entity, $form, $form_state);
  }

  /**
   * Copies form values, or a subset, to entity properties.
   *
   * This should not change existing entity properties that are not being edited
   * by this form.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity the current form should operate upon.
   * @param array $form
   *   A nested array of form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param string $validate_component
   *   (Optional) Only make sure the basic properties of the specified component
   *   are updated, in order to be able to validate the full component's config.
   *   Do not copy the component's full (unvalidated) configuration yet; leave
   *   other component's values in an undefined state.
   */
  protected function doCopyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state, string $validate_component = '') {
    assert($entity instanceof EntityCeDisplayInterface);
    $form_values = $form_state->getValues();
    if ($this->entity instanceof EntityWithPluginCollectionInterface) {
      // Do not manually update values represented by plugin collections.
      $form_values = array_diff_key($form_values, $this->entity->getPluginCollections());
    }

    // 'Rename' components, in a way that components can switch names.
    $this->renameEntityComponents($entity, $form, $form_state);

    // Collect data for 'regular' fields.
    // @todo Change this structure in #3446485: $form keys will be component
    //   names, not field names. Rename '#fields' (here and elsewhere)?
    foreach ($form['#fields'] as $field_name) {
      $values = $form_values['fields'][$field_name];
      $component_name = $values['name'];
      if ($validate_component && $validate_component !== $component_name) {
        continue;
      }

      if ($values['region'] == 'hidden') {
        $entity->removeComponent($field_name);
      }
      else {
        $options = $entity->getComponent($component_name);
        $options['formatter'] = $values['formatter'];
        $options['field_name'] = $field_name;
        $options['is_slot'] = (bool) $values['is_slot'];
        $options['weight'] = $values['weight'];
        $options['region'] = $values['region'];

        // Update field settings only if the submit handler told us to. Ignore
        // if we're being called to validate these settings.
        if ($form_state->get('plugin_settings_update') === $field_name && !$validate_component) {
          $component_name = $this->getComponentNameFromFieldName($field_name);
          // getRenderer() needs basic properties to be updated.
          $entity->setComponent($component_name, $options);
          $plugin = $entity->getRenderer($component_name);
          $plugin_form =& $form['fields'][$field_name]['plugin']['settings_edit_form']['form'];
          $subform_state = SubformState::createForSubform($plugin_form, $form, $form_state);
          $plugin->submitConfigurationForm($plugin_form, $subform_state);
          // Remove some fixed variables which are passed as part of
          // configuration from configuration storage.
          // @see \Drupal\custom_elements\Entity\EntityCeDisplay::getRenderer()
          $options['configuration'] = array_diff_key(
            $plugin->getConfiguration(),
            array_flip(['field_definition', 'view_mode', 'name', 'is_slot'])
          );
          $form_state->set('plugin_settings_update', NULL);
        }

        $entity->setComponent($component_name, $options);
      }
    }
    $entity->setCustomElementName($form_values['custom_element_name']);
    $entity->setForceAutoProcessing($form_values['force_auto']);
  }

  /**
   * Renames CE display components according to form state info.
   *
   * This method must be called exactly once on $this->entity, each time a full
   * form is submitted (and rebuilt on AJAX submit, or submitted/saved).
   * Calling this more than once
   * - may result in data corruption if a component is renamed to another
   *   existing component (if that component is also renamed at the same time;
   *   otherwise a validation error occurs and this is supposedly not called).
   * - will result in warnings (but no data corruption), otherwise. (We can
   *   therefore assume that data corruption never occurs if no warnings are
   *   ever logged.)
   *
   * @param \Drupal\custom_elements\Entity\EntityCeDisplayInterface $entity
   *   The CE display whose components to rename. The component names must be
   *   equal to the 'existing_name' form state values, and will get new names
   *   if applicable.
   * @param array $form
   *   A nested array of form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @todo This setup fails when doing any editing after a rename. It should
   *   probably be replaced by a setup that keeps the original name in
   *   $this->entity (with all other changed config) and only does the renaming
   *   on 'real' submit.
   */
  protected function renameEntityComponents(EntityCeDisplayInterface $entity, array $form, FormStateInterface $form_state) {
    // 'Rename' components, in a way that components can switch names.
    $new_components = [];
    $form_values = $form_state->getValues();
    // @todo Change this structure in #3446485: $form keys will be component
    //   names, not field names. Rename '#fields' (here and elsewhere)?
    foreach ($form['#fields'] as $field_name) {
      $values = $form_values['fields'][$field_name];
      $component_name = $values['name'];
      if ($values['region'] != 'hidden') {
        if (isset($values['parent_wrapper']['existing_name'])) {
          $existing_name = $values['parent_wrapper']['existing_name'];
          if ($existing_name && $component_name !== $existing_name) {
            // Old component can be empty (and populated later) if it used to
            // be hidden.
            $new_components[$component_name] = $entity->getComponent($existing_name) ?? [];
            $entity->removeComponent($existing_name);
          }
        }
        else {
          // This should never happen. Are we being called twice?
          $this->logger('custom_elements')->warning('Field row for @field_name somehow has no existing-name. (This suggests it is being renamed more than once.)', ['@field_name' => $field_name]);
        }
      }
    }
    foreach ($new_components as $component_name => $component) {
      $entity->setComponent($component_name, $component);
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function getDisplayModes() {
    return $this->entityDisplayRepository->getViewModes($this->entity->getTargetEntityTypeId());
  }

  /**
   * {@inheritdoc}
   */
  protected function getDisplayModeOptions() {
    return $this->entityDisplayRepository->getViewModeOptions($this->entity->getTargetEntityTypeId());
  }

  /**
   * {@inheritdoc}
   */
  protected function getDisplayModesLink() {
    return [
      '#type' => 'link',
      '#title' => $this->t('Manage view modes'),
      '#url' => Url::fromRoute('entity.entity_view_mode.collection'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function getOverviewUrl($mode) {
    $entity_type = $this->entityTypeManager->getDefinition($this->entity->getTargetEntityTypeId());
    return Url::fromRoute('entity.entity_ce_display.' . $this->entity->getTargetEntityTypeId() . '.view_mode', [
      'view_mode_name' => $mode,
    ] + FieldUI::getRouteBundleParameter($entity_type, $this->entity->getTargetBundle()));
  }

  /***
   * Allows all fields and not only whose display is configurable.
   *
   * @return \Drupal\Core\Field\FieldDefinitionInterface[]
   *   The array of field definitions
   */
  protected function getFieldDefinitions() {
    return $this->entityFieldManager->getFieldDefinitions($this->entity->getTargetEntityTypeId(), $this->entity->getTargetBundle());
  }

  /**
   * {@inheritdoc}
   */
  protected function getApplicablePluginOptions(FieldDefinitionInterface $field_definition) {
    $options = $this->customElementFormatterPluginManager->getOptions($field_definition->getType());
    $applicable_options = [];
    foreach ($options as $plugin_id => $label) {
      $plugin_class = DefaultFactory::getPluginClass($plugin_id, $this->customElementFormatterPluginManager->getDefinition($plugin_id));
      try {
        if ($plugin_class::isApplicable($plugin_id, $field_definition)) {
          $applicable_options[$plugin_id] = $label;
        }
      }
      catch (\Exception $exception) {
        $this->logger('custom_elements')
          ->warning('Error when determining compatible plugins: @error', ['@error' => $exception->getMessage()]);
      }
    }
    return $applicable_options;
  }

  /**
   * Returns an array containing the table headers.
   *
   * @return array
   *   The table header.
   */
  protected function getTableHeader() {
    return [
      $this->t('Field'),
      $this->t('Key'),
      $this->t('Is Slot'),
      $this->t('Formatter'),
      $this->t('Weight'),
      // There is a hidden column for this "parent_wrapper" element.
      '',
      ['data' => $this->t('Display')],
      ['data' => $this->t('Settings'), 'colspan' => 3],
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function thirdPartySettingsForm(PluginSettingsInterface $plugin, FieldDefinitionInterface $field_definition, array $form, FormStateInterface $form_state) {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  protected function alterSettingsSummary(array &$summary, PluginSettingsInterface $plugin, FieldDefinitionInterface $field_definition) {}

}
