<?php

namespace Drupal\custom_elements\Entity;

use Drupal\Core\Entity\EntityDisplayBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinition;
use Drupal\Core\Plugin\DefaultLazyPluginCollection as DefaultLazyPluginCollectionAlias;
use Drupal\custom_elements\CustomElementGenerator;

/**
 * Custom element display configuration entity.
 *
 * Contains custom element display options for all components of a rendered
 * entity in a given view mode.
 *
 * @ConfigEntityType(
 *   id = "entity_ce_display",
 *   label = @Translation("Entity custom element display"),
 *   entity_keys = {
 *     "id" = "id",
 *     "status" = "status"
 *   },
 *   handlers = {
 *     "access" = "\Drupal\custom_elements\Entity\Access\EntityCeDisplayAccessControlHandler",
 *   },
 *   config_export = {
 *     "id",
 *     "targetEntityType",
 *     "bundle",
 *     "mode",
 *     "customElementName",
 *     "forceAutoProcessing",
 *     "content",
 *     "hidden",
 *   }
 * )
 */
class EntityCeDisplay extends EntityDisplayBase implements EntityCeDisplayInterface {

  /**
   * {@inheritdoc}
   */
  protected $displayContext = 'view';

  /**
   * Whether this display is enabled or not.
   *
   * If the entity (form) display is disabled, we'll fall back to the 'default'
   * display.
   *
   * @var bool
   */
  protected $status = TRUE;

  /**
   * Custom element name to be displayed.
   *
   * @var string
   */
  protected $customElementName;

  /**
   * Status for 'ignore this config / fall back to using processors'.
   *
   * @var bool
   */
  protected $forceAutoProcessing = FALSE;

  /**
   * Custom elements generator.
   *
   * @var \Drupal\custom_elements\CustomElementGenerator
   */
  protected CustomElementGenerator $ceGenerator;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $values, $entity_type) {
    $this->pluginManager = \Drupal::service('custom_elements.plugin.manager.field.custom_element_formatter');
    $this->ceGenerator = \Drupal::service('custom_elements.generator');
    parent::__construct($values, $entity_type);
  }

  /**
   * {@inheritdoc}
   */
  public function setOriginalMode(string $viewMode): self {
    $this->originalMode = $viewMode;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getForceAutoProcessing(): bool {
    return !empty($this->forceAutoProcessing);
  }

  /**
   * {@inheritdoc}
   */
  public function setForceAutoProcessing(bool $status): self {
    $this->forceAutoProcessing = $status;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getCustomElementName() {
    return $this->customElementName;
  }

  /**
   * {@inheritdoc}
   */
  public function setCustomElementName($name) {
    $this->set('customElementName', $name);
    return $this;
  }

  /**
   * Gets the formatter for a CE display component.
   *
   * @param string $field_name
   *   The name for the CE display component. NOTE this not the field name;
   *   it's only called $field_name because of the existing
   *   EntityDisplayInterfaceinterface. (Other implementations also effectively
   *   pass the 'component name', because the only components are fields.)
   * @param bool $get_actual_field_name
   *   (Optional) Actually DO treat $field_name as a field name. DO NOT USE
   *   THIS; it's a temporary measure to support 'hidden'/new field rows in the
   *   UI, until they are removed / redone as component rows, in #3446485.
   *
   * @return \Drupal\custom_elements\CustomElementsFieldFormatterInterface|null
   *   A formatter plugin or NULL if the component does not exist.
   *
   * @todo remove $get_actual_field_name in #3446485.
   */
  public function getRenderer($field_name, bool $get_actual_field_name = FALSE) {
    if (isset($this->plugins[$field_name])) {
      return $this->plugins[$field_name];
    }
    $component = NULL;
    if (!$get_actual_field_name) {
      $component_name = $field_name;
      $component = $this->getComponent($component_name);
    }

    // Instantiate the formatter object from the stored display properties.
    $formatter = NULL;
    if ($component && isset($component['formatter'])) {
      // @todo When implementing static properties per #3446287, this code
      //   should know not to get a field name/definition, and the plugin
      //   manager should return something that is not a
      //   CustomElementsFieldFormatterInterface. Distinguish by the
      //   'formatter' property having a prefix + colon.
      // This is the actual field name.
      if (!isset($component['field_name'])) {
        // @todo Improve logging?
        return NULL;
      }
      $definition = $this->getFieldDefinition($component['field_name']);
      if ($definition) {
        $component += ['configuration' => [], 'name' => $component_name, 'is_slot' => FALSE];
        $formatter = $this->pluginManager->createInstance($component['formatter'], [
          'field_definition' => $definition,
          'view_mode' => $this->originalMode,
          'name' => $component['name'],
          'is_slot' => $component['is_slot'],
        ] + $component['configuration']
        );
      }
    }

    // Persist the formatter object.
    $this->plugins[$field_name] = $formatter;
    return $formatter;
  }

  /**
   * {@inheritdoc}
   */
  public function getPluginCollections() {
    $configurations = [];
    foreach ($this->getComponents() as $component_name => $component) {
      // @todo see getRenderer() when implementing #3446287; if() may need to
      //   change.
      if (isset($component['field_name'])) {
        $field_name = $component['field_name'];
        if (!empty($component['formatter']) && ($field_definition = $this->getFieldDefinition($field_name))) {
          $component += ['configuration' => [], 'name' => $component_name, 'is_slot' => FALSE];
          $configurations[$field_name] = [
            'id' => $component['formatter'],
            'field_definition' => $field_definition,
            'view_mode' => $this->originalMode,
            'name' => $component['name'],
            'is_slot' => $component['is_slot'],
          ] + $component['configuration'];
        }
      }
    }
    return [
      'formatters' => new DefaultLazyPluginCollectionAlias($this->pluginManager, $configurations),
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function init() {
    // Do not call parent::init(); we don't need its display defaults.
    if (!$this->targetEntityType) {
      return;
    }

    if (!$this->getCustomElementName()) {
      $custom_element = $this->ceGenerator->getViewModeDefaults($this->targetEntityType, $this->bundle, $this->originalMode);
      $this->setCustomElementName($custom_element->getPrefixedTag());
    }

    $initialized = !empty($this->content) || $this->forceAutoProcessing;
    if (!$initialized) {
      // Enable components options as done in the regular entity_view_display.
      /** @var \Drupal\Core\Entity\Display\EntityViewDisplayInterface $entity_view_display */
      $entity_view_display = \Drupal::service('entity_display.repository')
        ->getViewDisplay($this->targetEntityType, $this->bundle, $this->originalMode);
      $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions($this->targetEntityType, $this->bundle);

      // @todo Add support for statically set values.
      // Enable every component with "auto" that is enabled in the display.
      foreach ($entity_view_display->getComponents() as $field_name => $component) {
        // Ignore extra-fields.
        if (isset($field_definitions[$field_name])) {
          $component_name = str_starts_with($field_name, 'field_') ? substr($field_name, strlen('field_')) : $field_name;
          $this->setComponent($component_name, [
            'field_name' => $field_name,
            'formatter' => 'auto',
            'weight' => $component['weight'],
            'region' => $this->getDefaultRegion(),
            'is_slot' => str_starts_with($field_definitions[$field_name]->getType(), 'text') ? 1 : 0,
          ]);
        }
      }
    }

    $this->convertAlpha1Format();
  }

  /**
   * Converts pre-alpha2 format. Temporary backward compatibility code.
   *
   * @deprecated in custom_elements:3.0.0-alpha2 and is removed from
   *   custom_elements:3.0.0-alpha3. May or may not be alpha4 or later but
   *   likely before beta1.
   * @see https://www.drupal.org/project/custom_elements/issues/3455435
   *
   * @todo remove when implementing static properties per #3446287 or earlier.
   */
  protected function convertAlpha1Format(): void {
    // Either all components should be clearly old-format, or we exit silently.
    $components = $this->getComponents();
    if (!$components) {
      return;
    }
    foreach ($components as $component_name => $component) {
      if (isset($component['field_name']) || !isset($component['name'])) {
        return;
      }
    }

    // Always log, for as long as this method exists.
    \Drupal::logger('custom_elements')->notice('Converted CE display @id from old format. Please change code / saved config.', ['@id' => $this->id() . ' ' . json_encode($component)]);
    $this->content = [];
    foreach ($components as $field_name => $component) {
      $component_name = $component['name'];
      unset($component['name']);
      $component['field_name'] = $field_name;
      $this->setComponent($component_name, $component);
    }
    unset($this->hidden);
  }

  /**
   * {@inheritdoc}
   */
  public function setComponent($name, array $options = []) {
    // If no weight specified, make sure the field sinks at the bottom.
    if (!isset($options['weight'])) {
      $max = $this->getHighestWeight();
      $options['weight'] = isset($max) ? $max + 1 : 0;
    }
    $this->content[$name] = $options;
    unset($this->plugins[$name]);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function removeComponent($name) {
    unset($this->content[$name]);
    unset($this->plugins[$name]);

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    // Work around some of the parent code.
    $this->hidden = [];
    parent::preSave($storage);
    unset($this->hidden);
  }

  /**
   * Gets the definitions of the fields that are candidate for display.
   */
  protected function getFieldDefinitions() {
    if (!isset($this->fieldDefinitions)) {
      $definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions($this->targetEntityType, $this->bundle);

      $this->fieldDefinitions = $definitions;
      foreach ($this->fieldDefinitions as $field_definition) {
        if ($field_definition instanceof BaseFieldDefinition || $field_definition instanceof FieldDefinition) {
          // We'd like to be able to configure all the fields thus override
          // displayConfigurable to true for all field definitions.
          $field_definition->setDisplayConfigurable($this->displayContext, TRUE);
        }
      }
    }

    return $this->fieldDefinitions;
  }

}
